<?php
/**
 * Created by PhpStorm.
 * User: jderosa
 * Date: 21/08/19
 * Time: 11.30
 */

namespace FiloBlu\Rma\Helper;

use Exception;
use FiloBlu\Rma\Api\Data\EasyReturnInterface;
use FiloBlu\Rma\Api\RmaExtensionAttributesRepositoryInterface;
use FiloBlu\Rma\Logger\Logger;
use FiloBlu\Rma\Model\Carrier\Matrixrate;
use FiloBlu\Rma\Model\Carrier\MatrixrateFactory;
use FiloBlu\Rma\Model\Payment\ChangeSizePay;
use Magento\Catalog\Model\ProductRepository;
use Magento\Customer\Api\CustomerRepositoryInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Event\ManagerInterface;
use Magento\Framework\Exception\AlreadyExistsException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Model\AbstractExtensibleModel;
use Magento\Quote\Model\Quote\Address\Rate;
use Magento\Quote\Model\Quote\Address\RateRequestFactory;
use Magento\Quote\Model\QuoteFactory;
use Magento\Quote\Model\QuoteManagement;
use Magento\Quote\Model\ResourceModel\Quote;
use Magento\Rma\Api\Data\ItemInterface;
use Magento\Rma\Api\Data\RmaInterface;
use Magento\Rma\Api\RmaRepositoryInterface;
use Magento\Rma\Model\ResourceModel\Item;
use Magento\Rma\Model\Rma\Source\Status;
use Magento\Sales\Api\Data\OrderAddressInterfaceFactory;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\Data\OrderItemInterface;
use Magento\Sales\Api\OrderItemRepositoryInterface;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Model\Order;
use Magento\Sales\Model\Order\Email\Sender\OrderSender;
use Magento\Store\Model\StoreManagerInterface;
use Psr\Log\LoggerInterface;
use Throwable;

class ChangeSizeOrder
{

    /**
     * @var ProductRepository
     */
    protected $productRepository;
    /**
     * @var QuoteManagement
     */
    protected $quoteManagement;

    /**
     * @var CustomerRepositoryInterface
     */
    protected $customerRepository;

    /**
     * @var Rate
     */
    protected $shippingRate;

    /**
     * @var OrderItemRepositoryInterface
     */
    protected $orderItemRepositoryInterface;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var OrderRepositoryInterface
     */
    protected $orderInterface;

    /**
     * @var QuoteFactory
     */
    protected $quoteFactory;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var RmaRepositoryInterface
     */
    protected $rmaRepository;

    /**
     * @var RmaExtensionAttributesRepositoryInterface
     */
    protected $rmaExtensionAttributesRepository;

    /**
     * @var OrderSender
     */
    protected $orderSender;
    /**
     * @var Matrixrate
     */
    protected $matrixrate;
    /**
     * @var RateRequestFactory
     */
    protected $rateRequest;
    /**
     * @var array
     */
    protected $orderData;
    /**
     * @var int
     */
    protected $shippingAmount = 0;
    /**
     * @var int
     */
    protected $shippingAmountExclTax = 0;
    /**
     * @var int
     */
    protected $shippingTax = 0;
    /**
     * @var bool
     */
    protected $hasOrderPickupInStore = false;
    /**
     * @var int
     */
    protected $pickupInStoreAmount = 0;
    /**
     * @var string
     */
    protected $parentOrderShippingMethod;
    /**
     * @var string
     */
    protected $parentOrderShippingDescription;
    /**
     * @var array
     */
    protected $additionalInfo;
    protected $orderedProducts = [];
    /**
     * @var EasyReturnInterface
     */
    protected $easyReturn;
    /**
     * @var OrderAddressInterfaceFactory
     */
    private $addressInterfaceFactory;
    /**
     * @var Order
     */
    private $order;
    /**
     * @var Quote
     */
    private $quoteResource;
    /**
     * @var StoreManagerInterface
     */
    private $storeManager;
    /**
     * @var Item
     */
    private $itemResourceModel;
    /**
     * @var ChangeSizePay
     */
    private $changeSizePay;
    /**
     * @var Logger
     */
    private $filobluLogger;
    /**
     * @var ChangeSizeAttributeHelper
     */
    private $changeSizeAttributeHelper;
    /**
     * @var ManagerInterface
     */
    private $eventManager;

    /**
     * @var string
     */
    private $rateRequestType;
    /**
     * @var RmaHelper
     */
    private $rmaHelper;

    /**
     * ChangeSizeOrder constructor.
     * @param ScopeConfigInterface $scopeConfig
     * @param ProductRepository $productRepository
     * @param QuoteManagement $quoteManagement
     * @param CustomerRepositoryInterface $customerRepository
     * @param Rate $shippingRate
     * @param OrderItemRepositoryInterface $orderItemRepositoryInterface
     * @param OrderRepositoryInterface $orderInterface
     * @param QuoteFactory $quoteFactory
     * @param LoggerInterface $logger
     * @param RmaRepositoryInterface $rmaRepository
     * @param RmaExtensionAttributesRepositoryInterface $rmaExtensionAttributesRepository
     * @param OrderSender $orderSender
     * @param OrderAddressInterfaceFactory $addressInterfaceFactory
     * @param Order $order
     * @param Quote $quoteResource
     * @param StoreManagerInterface $storeManager
     * @param Item $itemResourceModel
     * @param ChangeSizePay $changeSizePay
     * @param Logger $filobluLogger
     * @param ChangeSizeAttributeHelper $changeSizeAttributeHelper
     * @param ManagerInterface $eventManager
     * @param MatrixrateFactory $matrixrate
     * @param RateRequestFactory $rateRequest
     * @param EasyReturnInterface $easyReturn
     * @param RmaHelper $rmaHelper
     */
    public function __construct(
        ScopeConfigInterface $scopeConfig ,
        ProductRepository $productRepository ,
        QuoteManagement $quoteManagement ,
        CustomerRepositoryInterface $customerRepository ,
        Rate $shippingRate ,
        OrderItemRepositoryInterface $orderItemRepositoryInterface ,
        OrderRepositoryInterface $orderInterface ,
        QuoteFactory $quoteFactory ,
        LoggerInterface $logger ,
        RmaRepositoryInterface $rmaRepository ,
        RmaExtensionAttributesRepositoryInterface $rmaExtensionAttributesRepository ,
        OrderSender $orderSender ,
        OrderAddressInterfaceFactory $addressInterfaceFactory ,
        Order $order ,
        Quote $quoteResource ,
        StoreManagerInterface $storeManager ,
        Item $itemResourceModel ,
        ChangeSizePay $changeSizePay ,
        Logger $filobluLogger ,
        ChangeSizeAttributeHelper $changeSizeAttributeHelper ,
        ManagerInterface $eventManager ,
        MatrixrateFactory $matrixrate ,
        RateRequestFactory $rateRequest,
        EasyReturnInterface $easyReturn,
        RmaHelper $rmaHelper
    )
    {
        $this->productRepository = $productRepository;
        $this->quoteManagement = $quoteManagement;
        $this->customerRepository = $customerRepository;
        $this->shippingRate = $shippingRate;
        $this->orderItemRepositoryInterface = $orderItemRepositoryInterface;
        $this->orderInterface = $orderInterface;
        $this->scopeConfig = $scopeConfig;
        $this->quoteFactory = $quoteFactory;
        $this->logger = $logger;
        $this->rmaRepository = $rmaRepository;
        $this->rmaExtensionAttributesRepository = $rmaExtensionAttributesRepository;
        $this->orderSender = $orderSender;
        $this->addressInterfaceFactory = $addressInterfaceFactory;
        $this->order = $order;
        $this->quoteResource = $quoteResource;
        $this->storeManager = $storeManager;
        $this->itemResourceModel = $itemResourceModel;
        $this->changeSizePay = $changeSizePay;
        $this->filobluLogger = $filobluLogger;
        $this->changeSizeAttributeHelper = $changeSizeAttributeHelper;
        $this->eventManager = $eventManager;
        $this->matrixrate = $matrixrate;
        $this->rateRequest = $rateRequest;
        $this->easyReturn = $easyReturn;
        $this->rmaHelper = $rmaHelper;
    }


    /**
     * @param array $orderData
     * @return bool|AbstractExtensibleModel|OrderInterface|object|null
     * @throws \Exception
     */
    public function createMageOrder(array $orderData )
    {
        $this->orderData = $orderData;


        /**
         * @var Order $order
         */
        $order = $orderData['order'];

        $request = $orderData['request'];
        if (!isset($request['filoblu_easy_return']) || $request['filoblu_easy_return'] != 'on') {
            if ($order->getShippingAddress()) {
                $request = $order->getShippingAddress()->toArray();
            } elseif ($order->getAddresses()) {
                $addresses = $order->getAddresses();
                if (count($addresses) == 1) {
                    $request = $addresses;
                } else {
                    $request = $addresses[0];
                    foreach ($addresses as $address) {
                        if ($address['address_type'] == 'shipping') {
                            $request = $address;
                            break;
                        }
                    }
                }
            }
        }

        try {
            // fix for api v1/returns POST that doesn't set Store
            if (!$this->storeManager->getStore()->getId() != $order->getStore()->getId()) {
                /**
                 * TODO: check order store id
                 */
                $this->storeManager->setCurrentStore($order->getStore());
            }

            // get original order store
            $orderStore = $order->getStore()->getWebsiteId();

            $orderSubmitted = $this->generateQuote($order , $orderStore , $request);
            //$orderSubmitted = $this->generateQuote2($order, $orderStore, $request);

            $new_order = $this->generateNewOrder($orderSubmitted , $order);

            $this->orderSender->send($new_order);

            return $new_order->getId();
        } catch (Exception $exception) {
            $this->filobluLogger->debug($exception->getMessage());
            throw new \Exception($exception->getMessage());
            //return false;
        } catch (Throwable $throwable) {
            $this->filobluLogger->debug($throwable->getMessage());
            throw new \Exception($throwable->getMessage());
            //return false;
        }

    }

    /**
     * @throws AlreadyExistsException
     * @throws LocalizedException
     */
    public function generateQuote(Order $order , $websiteId , $request)
    {
        // init quote
        $quote = $this->quoteFactory->create();
        $quote->setStore($order->getStore());
        $isCustomer = true;

        $_customer = null;

        try {
            $_customer = $this->customerRepository->get($order->getCustomerEmail() , $websiteId);
        } catch (Throwable $t) {
            $isCustomer = false;
        }

        // if you have already buyer id then you can load customer directly
        if (!$isCustomer) {
            //for guest order
            $quote->setCustomerIsGuest(true);

            $quote->setCustomerFirstname($request['firstname']);
            $quote->setCustomerLastname($request['lastname']);
            $quote->setCustomerEmail($order->getCustomerEmail());


        } else {
            $quote->assignCustomer($_customer); //Assign quote to customer
        }

        // $cart->setCurrency();
        //$quote->setCurrency($order->getOrderCurrency());

        // Set $returnAddress to quote shipping address
        $returnAddress = $this->addressInterfaceFactory->create();
        foreach ($request as $k => $val) {
            $returnAddress->setData($k , $val);
        }
        $quote->getShippingAddress()->addData((array)$returnAddress->getData());


        $quote->getBillingAddress()->addData((array)$order->getBillingAddress()->getData());

        $items = $this->getOrderDataByKey('items');

        $products = [];
        $orderItems = [];
        $final_price = 0;

        foreach ($items as $item) {
            /** @var OrderItemInterface $order_item */
            $orderItem = $this->orderItemRepositoryInterface->get($item['original_order_item_id']);
            $orderItems[] = $orderItem;
            $product = $this->productRepository->getById($item['product_id']);

            $final_price += $orderItem->getPriceInclTax();

            $qty = $item['qty'];

            $product->setStoreId($websiteId)->setPrice($final_price);
            $this->orderedProducts[$item['original_order_item_id']] = [
                'product' => $product ,
                'qty' => $qty
            ];
            $quote->addProduct($product , $qty);
        }

        //$final_price = $order->getBaseGrandTotal();

        //get original order shipping rate

        /*
        $rateRequest = $this->rateRequest->create()->addData($order->getShippingAddress()->toArray());
        $rateRequest->setWebsiteId($order->getStore()->getWebsiteId());
        $rateRequest->setDestCountryId($order->getShippingAddress()->getCountryId());
        $rateRequest->setData('order_total' , $order->getGrandTotal());
        $rateRequest->setData('tax_percentage' , $orderItems[0]->getTaxPercent());


        $originalShippingRate = $this->matrixrate->create()->collectRates($rateRequest);

        $this->rateRequestType = $rateRequest->getConditionMRName();

        $rates = $originalShippingRate->getAllRates();
        $rate = array_shift($rates);
        */

        $rate = $this->getRate($order, $orderItems);

        if ($rate->hasData('error_message')) {
            throw new \Exception('No matrix rate found!');
        }


        $this->shippingAmount = $rate->getPrice() ? $rate->getPrice() : $this->getRateAmount($request, $items);
        $this->shippingAmountExclTax = $rate->getCost() ? $rate->getCost() : $this->getRateAmount($request, $items, 'cost');
        $this->shippingTax = $this->shippingAmount - $this->shippingAmountExclTax;
        $this->parentOrderShippingDescription = 'Metodo di spedizione - Standard';
        $this->parentOrderShippingMethod = 'freeshipping_freeshipping';

        if ($this->rateRequestType == 'package_value') {
            $this->parentOrderShippingMethod = $rate->getMethod(); /** TODO: prevedere  */
        }

        $this->shippingRate->setCode($this->parentOrderShippingMethod);

        if ($order->getShippingMethod() == 'pickupinstore_pickupinstore') {
            $this->parentOrderShippingMethod = $order->getShippingMethod();
            $this->parentOrderShippingDescription = $order->getShippingDescription();
            $this->hasOrderPickupInStore = true;
            $this->pickupInStoreAmount = $order->getShippingAmount();
        }

        $shippingAddress = $quote->getShippingAddress();

        $shippingAddress->setCollectShippingRates(true)
            ->collectShippingRates()
            ->setShippingMethod($this->parentOrderShippingMethod);

        $quote->getShippingAddress()->addShippingRate($this->shippingRate);
        $quote->setPaymentMethod(ChangeSizePay::PAYMENT_CODE); //payment method

        $this->quoteResource->save($quote);

        $parentSkuReplaced = $this->getOrderDataByKey('parent_product_sku');

        $orderData = $this->getOrderDataByKey();

        $simpleItemIds = [];

        $rmaEntityId = 0;

        if ($orderData) {
            foreach ($order->getItems() as $orderItems) {

                if (!$orderItems->getParentItemId()) {
                    continue;
                }

                $key = $this->searchOrderItemKeyByValue($orderItems->getParentItemId());

                if ($key !== false) {
                    $simpleItemIds[] = $orderItems->getItemId();

                    if (!$rmaEntityId) {
                        $rmaEntityId = $items[$key]['rma_entity_id'];
                    }
                }
            }
        }


        $this->additionalInfo = [
            'parent_order_id' => $order->getId() ,
            'parent_order_increment_id' => $order->getIncrementId() ,
            'rma_id' => $orderData['rma_id'] ,
            'rma_entity_id' => $rmaEntityId ,
            'parent_order_item_id' => implode(',' , $simpleItemIds) ,
            'parent_product_sku' => $parentSkuReplaced ,
            'grand_total' => $final_price ,
            'base_grand_total' => $final_price
        ];


        $quote->getPayment()->setAdditionalInformation($this->additionalInfo);
        $quote->getPayment()->importData(['method' => ChangeSizePay::PAYMENT_CODE]);

        //prevent customer null when customer_is_guest
        if ($quote->getCustomerEmail() == null) {
            //for guest order
            $quote->setCustomerIsGuest(true);
            $quote->setCustomerFirstname($request['firstname']);
            $quote->setCustomerLastname($request['lastname']);
            $quote->setCustomerEmail($order->getCustomerEmail());
        }

        foreach ($quote->getAllItems() as $key => $quoteItem) {
            //$quote->getItemsCollection()->addItem($quoteItem);
            $this->createQuoteItem($quoteItem , $quote , $key);
        }

        $quote->collectTotals();

        $this->quoteResource->save($quote);

        return $this->quoteManagement->submit($quote);
    }

    /**
     * @param OrderInterface $order
     * @param OrderItemInterface[] $originalOrderItems
     * @return mixed|null
     */
    public function getRate(OrderInterface $order, array $originalOrderItems)
    {
        $taxPercent = 0.0;

        foreach ($originalOrderItems as $item) {
            if ((float)$item->getTaxPercent() > $taxPercent) {
                $taxPercent = (float)$item->getTaxPercent();
            }
        }

        $rateRequest = $this->rateRequest->create()->addData($order->getShippingAddress()->toArray());
        $rateRequest->setWebsiteId($order->getStore()->getWebsiteId());
        $rateRequest->setDestCountryId($order->getShippingAddress()->getCountryId());
        $rateRequest->setData('order_total' , $order->getGrandTotal());
        $rateRequest->setData('tax_percentage' , $taxPercent);


        $originalShippingRate = $this->matrixrate->create()->collectRates($rateRequest);

        $this->rateRequestType = $rateRequest->getConditionMRName();

        $rates = $originalShippingRate->getAllRates();

        if (empty($rates)) {
            throw new \RuntimeException('Could not find a rate for order');
        }

        return  array_shift($rates);
    }

    public function getRateAmount($request, $items, $type = 'price') {

        $rateData = array_merge($request, [
            'add_currency' => false,
            'price_type' => $type,
            'items' => $items
        ]);

        $rateData['items'] = array_map(function(&$item) {
            $item['order_item_id'] = $item['original_order_item_id'];
            $item['qty_requested'] = (float)$this->orderedProducts[$item['original_order_item_id']];
            return $item;
        }, $rateData['items']);

        return (float)$this->easyReturn->estimateReturnShippingCost($rateData);
    }

    /**
     * @throws \Exception
     */
    public function getOrderDataByKey($key = '')
    {

        if (!isset($this->orderData)) {
            throw new Exception('Order Data is empty!');
        }

        if (!$key) {
            return $this->orderData;
        }

        if (!isset($this->orderData[$key])) {
            throw new Exception(sprintf('Order Data key "%s" des not exists' , $key));
        }

        return $this->orderData[$key];
    }

    public function searchOrderItemKeyByValue($value , $key = 'original_order_item_id')
    {

        $items = $this->getOrderDataByKey('items');

        return array_search($value , array_column($items , $key));
    }

    public function createQuoteItem(\Magento\Quote\Model\Quote\Item $quoteItem , $quote)
    {

        $originalOrder = $this->getOrderDataByKey('order');

        $itemKey = $this->searchOrderItemKeyByValue($quoteItem->getProductId() , 'product_id');
        $items = $this->getOrderDataByKey('items');
        $originalOrderItemId = $items[$itemKey]['original_order_item_id'];

        $qty = $this->orderedProducts[$originalOrderItemId]['qty'];

        $originalItem = $originalOrder->getItems()[$originalOrderItemId];
        $quoteItem->setPrice($originalItem->getPrice() * $qty);
        $quoteItem->setOriginalPrice($originalItem->getOriginalPrice() * $qty);
        $quoteItem->setDiscountAmount(((float)$originalItem->getDiscountAmount() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setBaseDiscountAmount(((float)$originalItem->getBaseDiscountAmount() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setDiscountPercent($originalItem->getDiscountPercent());
        $quoteItem->setBasePrice($originalItem->getBasePrice() * $qty);
        $quoteItem->setBaseOriginalPrice($originalItem->getBaseOriginalPrice() * $qty);
        $quoteItem->setTaxAmount((float)$originalItem->getTaxAmount() * $qty / (int)$originalItem->getQtyOrdered());
        $quoteItem->setBaseTaxAmount((float)$originalItem->getBaseTaxAmount() * $qty / (int)$originalItem->getQtyOrdered());
        $quoteItem->setBaseRowTotal(((float)$originalItem->getBaseRowTotal() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setBaseRowTotalInclTax(((float)$originalItem->getBaseRowTotalInclTax() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setRowTotal((((float)$originalItem->getRowTotal() / $originalItem->getQtyOrdered()) * $qty));
        $quoteItem->setRowTotalInclTax(((float)$originalItem->getRowTotalInclTax() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setPriceInclTax($originalItem->getPriceInclTax() * $qty);
        $quoteItem->setBasePriceInclTax($originalItem->getBasePriceInclTax() * $qty);
        $quoteItem->setDiscountCalculationPrice((float)$originalItem->getPriceInclTax() * $qty);
        $quoteItem->setBaseDiscountCalculationPrice((float)$originalItem->getPriceInclTax() * $qty);
        $quoteItem->setBaseGrandTotal((float)$originalItem->getPriceInclTax() * $qty);
        if ($originalItem->getDiscountAmount() != null) {
            $quoteItem->setDiscountDescription($this->getOrderDataByKey('order')->getDiscountDescription());
        }
        $quoteItem->setBaseDiscountAmount(($originalItem->getBaseDiscountAmount() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setDiscountTaxCompensationAmount(($originalItem->getDiscountTaxCompensationAmount() / $originalItem->getQtyOrdered()) * $qty);
        $quoteItem->setBaseDiscountTaxCompensationAmount(($originalItem->getBaseDiscountTaxCompensationAmount() / $originalItem->getQtyOrdered()) * $qty);


        //reset quote Items
        $quote->getItemsCollection()->removeItemByKey($itemKey);
        $quote->getItemsCollection()->addItem($quoteItem);
    }

    public function generateNewOrder($orderSubmitted , $order)
    {
        $orderData = $this->getOrderDataByKey();

        $items = $this->getOrderDataByKey('items');

        $new_order = $this->orderInterface->get($orderSubmitted->getId());

        $this->eventManager->dispatch(
            'filoblu_rma_after_change_size_create_order' ,
            [
                'new_order' => $new_order ,
                'parent_order' => $order
            ]
        );

        $rmaEntityId = 0;

        $subtotals = [
            'subtotal' => 0,
            'subtotalInclTax' => 0,
            'taxAmount' => 0
        ];

        $isLastItem = $orderData['is_last_item'];

        foreach ($items as $key => $item) {
            $qty = $item['qty'];
            $product = $this->productRepository->getById($item['product_id']);

            $originalOrder = $this->getOrderDataByKey('order');

            $itemKey = $this->searchOrderItemKeyByValue($product->getEntityId() , 'product_id');

            $originalOrderItemId = $items[$itemKey]['original_order_item_id'];
            $originalItem = $originalOrder->getItems()[$originalOrderItemId];

            $newOrderItem = $new_order->getItems()[$key];
            $productOptions = $newOrderItem->getProductOptions();
            $newOrderItem->setProductOptions(
                array_merge($productOptions ,
                    ['attributes_info' =>
                        [
                            [
                                'label' => $originalItem->getProductOptions()['attributes_info'][0]['label'] ,
                                'option_id' => $originalItem->getProductOptions()['attributes_info'][0]['option_id'] ,
                                'value' => $product->getResource()->getAttribute('size')->getSource()->getOptionText($product->getSize()) ,
                                'option_value' => $product->getSize()
                            ]
                        ]
                    ]
                )
            )->save();

            $subtotals['subtotal'] += $newOrderItem->getRowTotal();
            $subtotals['subtotalInclTax'] += $newOrderItem->getRowTotalInclTax();
            $subtotals['taxAmount'] += $newOrderItem->getBaseTaxAmount();

            $newOrderItem->setPrice((float)$originalItem->getBasePrice() * $qty);
            $newOrderItem->setRowTotalInclTax(((float)$originalItem->getRowTotalInclTax() / $originalItem->getQtyOrdered()) * $qty);
            $newOrderItem->save();

            if (!$rmaEntityId) {
                $rmaEntityId = $item['rma_entity_id'];
            }
        }

        $subtotals['taxAmount'] += $this->shippingTax;

        //setting new_order data according to parent order
        $new_order->setShippingDescription($this->parentOrderShippingDescription);
        // set pending status - emmedata sync
        $new_order->setStatus('pending');
        $new_order->save();
        // set change_size status
        $new_order->setStatus(ChangeSizeAttributeHelper::ORDER_STATUS_CHANGESIZE_CODE);
        $new_order->setSendEmail(true);
        $orderDetails = '';
        foreach ($this->additionalInfo as $key => $val) {
            $orderDetails .= "$key:$val";
            if (next($this->additionalInfo)) {
                $orderDetails .= ', ';
            }
        }
        $new_order->addStatusHistoryComment("This order is generated programatically . Some details: $orderDetails");

        $this->setNewOrderTotals($new_order, $order, $subtotals, $isLastItem);

        $new_order->save();

        $new_order = $this->order->loadByIncrementId($new_order->getIncrementId());

        //save changeSizeOrderId into rma
        $this->rmaExtensionAttributesRepository->setChangeSizeOrderId($rmaEntityId , $new_order->getEntityId());

        return $new_order;
    }

    public function setNewOrderTotals(\Magento\Sales\Model\Order $new_order, $order, $subtotals, $isLastItem) {

        if ($this->rmaHelper->getChangeSizeOrderCreationStrategy() == 'each' && !$isLastItem) {
            $this->clearShippingAmounts();
        }

        if ($this->isOrderTotallyReturned($order->getId())) {
            //change totals
            $new_order->setMspCodAmount($order->getMspCodAmount());
            $new_order->setBaseMspCodAmount($order->getBaseMspCodAmount());
            $new_order->setMspCodTaxAmount($order->getMspCodTaxAmount());
            $new_order->setBaseMspCodTaxAmount($order->getBaseMspCodTaxAmount());

            //calculate tax from original order
            if ($order->getGrandTotal() == 0) {
                $originalTaxPercent = 0;
            } else {
                $originalTaxPercent = $order->getBaseTaxAmount() / $order->getGrandTotal();
            }

            if (!$this->hasOrderPickupInStore) {
                $new_order->setShippingAmount($this->shippingAmountExclTax ?: $order->getShippingAmount());
                $new_order->setBaseShippingAmount($this->shippingAmountExclTax ?: $order->getBaseShippingAmount());

                $new_order->setShippingInclTax($this->shippingAmount ?: $order->getShippingInclTax());
                $new_order->setBaseShippingInclTax($this->shippingAmount ?: $order->getBaseShippingInclTax());
                $new_order->setBaseShippingTaxAmount($this->shippingTax ?: $order->getBaseShippingTaxAmount());
                $new_order->setShippingTaxAmount($this->shippingTax ?: $order->getShippingTaxAmount());
            } else {
                $new_order->setShippingAmount($this->shippingAmountExclTax ?: 0);
                $new_order->setBaseShippingAmount($this->shippingAmountExclTax ?: 0);

                $new_order->setShippingInclTax($this->shippingAmount ?: 0);
                $new_order->setBaseShippingInclTax($this->shippingAmount ?: 0);
                $new_order->setBaseShippingTaxAmount($this->shippingTax ?: 0);
                $new_order->setShippingTaxAmount($this->shippingTax ?: 0);
            }

            $new_order->setBaseGrandTotal($subtotals['subtotalInclTax'] + $this->shippingAmount);
            $new_order->setGrandTotal($subtotals['subtotalInclTax'] + $this->shippingAmount);

            $new_order->setBaseTotalPaid($subtotals['subtotalInclTax'] + $this->shippingAmount);
            $new_order->setSubtotal($subtotals['subtotal']);
            $new_order->setSubtotalInclTax($subtotals['subtotalInclTax']);
            $new_order->setBaseSubtotal($subtotals['subtotal']);
            $new_order->setBaseSubtotalInclTax($subtotals['subtotalInclTax']);
            $new_order->setTaxAmount($subtotals['taxAmount']);
            $new_order->setBaseTaxAmount($subtotals['taxAmount']);

            if ($order->getMspCodAmount()) {
                $feeTaxes = $originalTaxPercent * ($order->getMspCodAmount() + ($this->shippingAmount ?: $order->getShippingInclTax()));
                $new_order->setBaseTaxAmount((float)$feeTaxes + $new_order->getBaseTaxAmount());
                $new_order->setTaxAmount($new_order->getBaseTaxAmount());
            }
        }
    }

    public function clearShippingAmounts() {
        $this->shippingAmountExclTax = 0;
        $this->shippingTax = 0;
        $this->shippingAmount = 0;
    }

    /**
     * @param int $orderId
     * @return bool
     */
    public function isOrderTotallyReturned(int $orderId)
    {
        $itemResourceModel = $this->itemResourceModel;
        $returnableItems = $itemResourceModel->getReturnableItems($orderId);
        $fullRma = true;
        foreach ($returnableItems as $item) {
            if ($item > 0) {
                $fullRma = false;
                break;
            }
        }

        return $fullRma;
    }

    /**
     * @param $orderId
     * @return bool
     */
    public function isValidForUpdateStatus($orderId)
    {
        try {
            $order = $this->orderInterface->get($orderId);
            if ($order) {
                try {
                    $paymentInfo = $order->getExtensionAttributes()->getPaymentAdditionalInfo() ?? [];
                    $rmaId = 0;
                    foreach ($paymentInfo as $info) {
                        if ($info->getKey() && $info->getKey() == 'rma_id' && $info->getValue()) {
                            $rmaId = $info->getValue();
                            break;
                        }
                    }
                } catch (Throwable $t) {
                    $paymentInfo = $order->getPayment()->getAdditionalInformation() ?? [];
                    $rmaId = 0;
                    foreach ($paymentInfo as $k => $info) {
                        if ($k == 'rma_id' && $info) {
                            $rmaId = $info;
                            break;
                        }
                    }
                }


                $rma = $this->rmaRepository->get($rmaId);
                if (!$rma) {
                    return false;
                }
                $rmaStatus = $rma->getStatus();
                //case ($result->getStatus() == Status::STATE_PROCESSED_CLOSED || $result->getStatus() == Status::STATE_DENIED || $result->getStatus() == Status::STATE_CLOSED) :
                $allowedStatus = [Status::STATE_PROCESSED_CLOSED , Status::STATE_DENIED , Status::STATE_CLOSED , Status::STATE_APPROVED_ON_ITEM];
                if (in_array($rmaStatus , $allowedStatus)) {
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        } catch (Exception $e) {
            $this->filobluLogger->debug($e->getMessage());
            return false;
        } catch (Throwable $t) {
            $this->filobluLogger->debug($t->getMessage());
            return false;
        }
    }

    /**
     * @param $order
     * @return bool
     */
    public function isChangeSizeOrder($order)
    {
        return $this->changeSizeAttributeHelper->isChangeSizeOrder($order);
    }

    /**
     * @param \Magento\Rma\Api\Data\RmaInterface $rma
     * @return array
     */
    public function getChangeSizeItems(RmaInterface $rma) {

        $changeSizeItems = [];

        foreach ($rma->getItems() as $item) {
            if (!$item->getExtensionAttributes() || !$item->getExtensionAttributes()->getChangeSize()) {
                continue;
            }

            $changeSizeItems[] = $item;
        }

        return $changeSizeItems;
    }

    /**
     * @param RmaInterface $rma
     * @return array
     */
    public function getOrderItemsByRma(RmaInterface $rma): array
    {
        $rmaItems = $this->getChangeSizeItems($rma);

        if (empty($rmaItems)) {
            return [];
        }

        $orderItems = [];

        /** @var ItemInterface $rmaItem */
        foreach ($rmaItems as $rmaItem) {
            $orderItems[] = $this->orderItemRepositoryInterface->get($rmaItem->getOrderItemId());
        }

        return $orderItems;
    }
}
